import numpy as np

class Molecule:
    def __init__(self, elements, coordinates, charge=0, spin=0, name=''):
        self.elements = np.array(elements)
        self.coordinates = np.array(coordinates)
        self.charge = charge
        self.spin = spin
        self.name = name

    def get_num_atoms(self):
        return len(self.elements)
    
    def get_atom_coordinates(self,index):
        return self.coordinates[index-1][:]
    
    def add_atom(self, element, coordinate):
        self.elements = np.append(self.elements, element)
        self.coordinates = np.append(self.coordinates, coordinate).reshape(-1, 3)

    def set_charge(self,charge):
        self.charge = charge

    def set_spin(self,spin):
        self.spin = spin
    
    def set_name(self,name):
        self.name = name




import os
import time
from APIs.Molecule import Molecule
import numpy as np


'''
Calculators are:
    xTB: GFN1-xTB, GFN2-xTB
    ANI: ANI-1x, ANI-2x, ANI-1ccx
    MLatom: AIQM1, AIQM1@DFT
    Orca: 
'''
global eleslist
eleslist = ['H','He',
        'Li','Be','B','C','N','O','F','Ne',
        'Na','Mg','Al','Si','P','S','Cl','Ar',
        'K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr']

global b2a
b2a = 0.52917724

#XTB default input is in Bohr
def calculate_mol_xTB(mol,xtbpara):
    #first solve the environment
    try:
        from xtb.libxtb import VERBOSITY_FULL, VERBOSITY_MINIMAL
        from xtb.interface import Environment, Calculator, Param
    except ImportError:
        print('no xTB module found.')
    
    env = Environment()
    env.set_output("env_error.log")
    env.set_verbosity(VERBOSITY_FULL)
    if env.check != 0:
        env.show("Error message")
    env.release_output()
    xtbpara = Param(xtbpara)
    #define structure data for molecule
    eles_nums = mol.elements
    coors = np.divide(mol.coordinates,b2a)
    #define calculator based on input xtbpara and mol
    calc = Calculator(xtbpara,eles_nums,coors)
    calc.set_verbosity(VERBOSITY_MINIMAL)
    #Calculation
    res = calc.singlepoint()
    Ene = res.get_energy()
    Grad = res.get_gradient()
    return Ene, Grad

#ani input in cartesian(Angstrom)
def calculate_mol_ani(mol,methodani):
    
    #ANI
    try:
        import torch
        import torchani
        device = torch.device('cpu')

    except ImportError:
        print('no ANI module found')

    #define model based on input methodani

    if methodani.lower() == 'ani-2x':
        model = torchani.models.ANI2x(periodic_table_index=True).to(device).double()
    elif methodani.lower() == 'ani-1ccx':
        model = torchani.models.ANI1ccx(periodic_table_index=True).to(device).double()
    elif methodani.lower() == 'ani-1x':
        model = torchani.models.ANI1x(periodic_table_index=True).to(device).double()
    else:
        print('Error: Method not recognised, Using ANI2x instead')
        model = torchani.models.ANI2x(periodic_table_index=True).to(device).double()

    #have to convert from np to tensor first
    coordinates = torch.from_numpy(mol.coordinates).requires_grad_(True).unsqueeze(0)
    species = torch.from_numpy(mol.elements).unsqueeze(0)
    ene = model((species, coordinates)).energies
    derivative = torch.autograd.grad(ene.sum(), coordinates)[0]
    force = -derivative
    #convert tensor to numpy and return
    return ene.item(),derivative.numpy()


def calculate_mol_mlatom(xyzfile,charge,spin,methodmlatm):
    try:
        import mlatom as ml
    except ImportError:
        print('no MLatom module found')
    #read molecule in 
    mol = ml.data.molecule.from_xyz_file(xyzfile)
    mol.charge = charge
    mol.multiplicity = spin
    #define model
    model = ml.models.methods(method=methodmlatm)
    
    #calculate energy and force
    model.predict(molecule=mol)

    ene = mol.energy
    grad = mol.get_energy_gradients()
    return ene,grad

#OCRA: orca takes some file in&out without python API.
#So I just write a function to call orca and get the result.
def mol2xyz(mol):
    nums = mol.get_num_atoms()
    fw = open(f'{mol.name}.xyz','w')
    fw.write('%d\n\n'%(nums))
    try:
        for i in range(nums):
            fw.write('%s %f %f %f\n'%(eleslist[mol.elements[i]-1],mol.coordinates[i][0],mol.coordinates[i][1],mol.coordinates[i][2]))
    except IndexError:
            print('IndexError: Maybe no such element in eleslist')
    fw.write('\n')
    fw.close()        

def write_orca_input(mol,charge,spin,methodorca,par_core):
    fw = open(f'{mol.name}.inp','w')
    fw.write(f'! {methodorca} DEF2-TZVPP ENGRAD\n')
    fw.write(f'%PAL NPROCS {par_core} END\n')
    #Use xyz file to define molecule, it is covenient
    fw.write(f'* xyzfile {charge} {spin} {mol.name}.xyz \n')
    fw.write('\n')


def Get_Orca_SP(outname):
    fr = open(outname,"r") 
    lines = fr.readlines()
    fr.close()
    index_of_energy = []
    index_of_ext_energy = []
    i = 0
    errlines = []
    for line in lines:
        if "FINAL SINGLE POINT ENERGY" in line:
            index_of_energy.append(i)
        
        i= i+1
    try:
        loc = int(index_of_energy[-1])
        energy = lines[loc].split()[-1]
        return float(energy)           
    except IndexError:
        print('Energy not found in %s file!'%outname)              #help check the bug   
        print(errlines)
        print('')

def Get_orca_grad(outname,mol,der):
    fg = open(outname,"r")
    lines = fg.readlines()
    fg.close()
    nums = mol.get_num_atoms()
    grad_arr = np.zeros((nums,3),dtype=float)
    if der == 1:
        for line in lines:
            if "CARTESIAN GRADIENT" in line:
                index_of_grad = lines.index(line)
                break
            counts = 0
        for grads in lines[index_of_grad+3:index_of_grad+2+nums]:
            grad_arr[counts] = np.array([float(grads.split()[3]),float(grads.split()[4]),float(grads.split()[5])])
            counts += 1
        
    return grad_arr


def calculate_mol_orca(mol,charge,spin,methodorca,par_core,der):
    write_orca_input(mol,charge,spin,methodorca,par_core)
    orca_inp = f'{mol.name}.inp'
    os.system('orca '+orca_inp+'> '+mol.name+'.log')
    ene_sp = Get_Orca_SP(mol.name+'.log')
    grad_sp = Get_orca_grad(mol.name+'.log',mol,der)
    return ene_sp, grad_sp




from APIs.Molecule import Molecule
from APIs import New_Calculator
from sys import argv
import numpy as np
import os 



def get_external_coord(filein):
    with open(filein, 'r') as f:
        [atoms, deriva, charge ,spin] = [int(s) for s in f.readline().split()]
        ele = np.zeros((atoms,), dtype = int)
        
        atomlist = np.zeros((atoms,3), dtype = float)

        for i in range(atoms):
            axestr = f.readline().split()
            ele[i] = int(axestr[0])
            #convert from Bohr to Angstrom
            atomlist[i][0] = float(axestr[1])*0.52917724
            atomlist[i][1] = float(axestr[2])*0.52917724
            atomlist[i][2] = float(axestr[3])*0.52917724
    return ele, atomlist, deriva, charge, spin

def write_fileout(fileout, Ene, Gradient,der):
    with open(fileout, 'w') as f:
        f.write('%20.12E%20.12E%20.12E%20.12E\n'% (Ene,0.0,0.0,0.0)) 
        if der == 0:
            for i in range(Gradient.shape[0]):
                f.write('%20.12E%20.12E%20.12E\n'% (0.0,0.0,0.0))
                f.write('\n')
        elif der == 1:
            for i in range(Gradient.shape[0]):
                f.write('%20.12E%20.12E%20.12E\n'% (Gradient[i][0],Gradient[i][1],Gradient[i][2]))
                f.write('\n')
    return





if __name__ == '__main__':
    if len(argv) < 8:
        print("You should define a certain method used to calculate this layer\n e.g.: APIs gfn2-xtb\n ")
        print("If you want to use orca, you should method meanwhile define the core \n e.g.: APIs method core\n")
    elif len(argv) == 9 :
        
        method_use = argv[1]
        core = argv[2]
        filein = argv[4]
        fileout = argv[5]
    elif len(argv) == 8:
        method_use = argv[1]
        filein = argv[3]
        fileout = argv[4]
    molname = f'{method_use}-input'
    elelist, coors, deriva, charge, spin = get_external_coord(filein)
    Mole = Molecule(elelist,coors,charge,spin,molname)
    New_Calculator.mol2xyz(Mole)
    #use different calculators to calculate the energy and gradient according to method defined.
    
    
    if method_use.lower() in ['xtb', 'gfn2-xtb', 'gfn1-xtb']:
        #os.system('source activate Strain_VIZ')
        print('Using xtb-python')
        if method_use == 'xtb' or method_use == 'gfn2-xtb':
            meth_in = 1                                  # gfn2xtb=1 in Param
        elif method_use == 'gfn1-xtb':
            meth_in = 2                                 ## gfn2xtb=2 in Param
        Ene, Gradient = New_Calculator.calculate_mol_xTB(Mole,meth_in)
        print('xtb calculation finished')
    elif method_use.lower() in ['ani-1x', 'ani-2x','ani-1ccx']:
        print('Using ani')
        os.system('source activate Strain_VIZ')
        Ene, Gradient = New_Calculator.calculate_mol_ani(Mole,method_use)
        print('ani calculation finished')
    elif method_use.lower() in ['aiqm1', 'aiqm1@dft']:
        print('Using mlatom')
        method_use = method_use.upper()
        Ene,Gradient = New_Calculator.calculate_mol_mlatom(f'{Mole.name}.xyz',charge,spin,method_use)  
        print('mlatom finished')
    else:
        os.system('source /share/home/yanz/bin/env/OC51.env')
        print(f'External will try use orca to compute the energy and gradient in {method_use} level')
        Ene, Gradient = New_Calculator.calculate_mol_orca(Mole,charge,spin,method_use,core,deriva)

    write_fileout(fileout, Ene, Gradient,deriva)


        
    
    
    
    
    


from setuptools import setup, find_packages

setup(
    name='GauAPIs',
    version='2.0',
    packages=find_packages(),
    install_requires=[
        
    ],
    author='Sam Yunteng Liao',
    author_email='lmasgne@gmail.com',
    description='APIs for Gaussian running other calculators through External',
    license='MIT',
    platforms="linux",
    keywords='Gaussian--external python APIs',
    url='https://github.com/SamYToeL/APIs'
)